from __future__ import unicode_literals

import logging
from collections import OrderedDict

from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from django_giropay import settings as django_giropay_settings
from django_giropay.constants import SHOPPING_CART_TYPE
from django_giropay.models import GiropayTransaction

import requests

from django_giropay.utils import build_giropay_full_uri

logger = logging.getLogger(__name__)


class GiropayWrapper(object):
    interface_version = 'django_giropay_v{}'.format(django_giropay_settings.DJANGO_GIROPAY_VERSION)

    api_url = django_giropay_settings.GIROPAY_API_URL
    bankstatus_url = django_giropay_settings.GIROPAY_BANKSTATUS_URL
    transaction_start_url = django_giropay_settings.GIROPAY_TRANSACTION_START_URL
    issuer_url = django_giropay_settings.GIROPAY_ISSUER_URL

    auth = None

    def __init__(self, auth=None):
        super(GiropayWrapper, self).__init__()
        if getattr(settings, 'GIROPAY', False):
            if auth:
                self.auth = auth
            else:
                self.auth = {
                    'MERCHANT_ID': django_giropay_settings.GIROPAY_MERCHANT_ID,
                    'PROJECT_ID': django_giropay_settings.GIROPAY_PROJECT_ID,
                    'PROJECT_PASSWORD': django_giropay_settings.GIROPAY_PROJECT_PASSWORD,
                }

    def start_transaction(self, merchant_tx_id, amount, purpose,
        currency='EUR', shoppingCartType=SHOPPING_CART_TYPE.ANONYMOUS_DONATION, shipping_address=None,
        merchantOrderReferenceNumber=False, kassenzeichen=False,
        redirect_url=django_giropay_settings.GIROPAY_RETURN_URL,
        notify_url=django_giropay_settings.GIROPAY_NOTIFICATION_URL,
        success_url=django_giropay_settings.GIROPAY_SUCCESS_URL,
        error_url=django_giropay_settings.GIROPAY_ERROR_URL
    ):
        giropay_transaction = GiropayTransaction()
        giropay_transaction.merchant_id = self.auth['MERCHANT_ID']
        giropay_transaction.project_id = self.auth['PROJECT_ID']
        giropay_transaction.merchant_tx_id = merchant_tx_id
        giropay_transaction.amount = amount
        giropay_transaction.currency = currency
        giropay_transaction.purpose = purpose
        giropay_transaction.redirect_url = build_giropay_full_uri(redirect_url)
        giropay_transaction.notify_url = build_giropay_full_uri(notify_url)
        giropay_transaction.success_url = build_giropay_full_uri(success_url)
        giropay_transaction.error_url = build_giropay_full_uri(error_url)

        # Ordering needs to be the same as in the API docs, otherwise the generated hash will be invalid
        data = OrderedDict()
        data['merchantId'] = self.auth['MERCHANT_ID']
        data['projectId'] = self.auth['PROJECT_ID']
        data['merchantTxId'] = merchant_tx_id
        data['amount'] = amount
        data['currency'] = currency
        data['purpose'] = purpose
        data['shoppingCartType'] = shoppingCartType
        if shipping_address:
            data['shippingAddresseFirstName'] = shipping_address.get('shippingAddresseFirstName', '')
            data['shippingAddresseLastName'] = shipping_address.get('shippingAddresseLastName', '')
            data['shippingCompany'] = shipping_address.get('shippingCompany', '')
            data['shippingAdditionalAddressInformation'] = shipping_address.get('shippingAdditionalAddressInformation',
                                                                                '')
            data['shippingStreet'] = shipping_address.get('shippingStreet', '')
            data['shippingStreetNumber'] = shipping_address.get('shippingStreetNumber', '')
            data['shippingZipCode'] = shipping_address.get('shippingZipCode', '')
            data['shippingCity'] = shipping_address.get('shippingCity', '')
            data['shippingCountry'] = shipping_address.get('shippingCountry', '')
            data['shippingEmail'] = shipping_address.get('shippingEmail', '')
        if merchantOrderReferenceNumber:
            data['merchantOrderReferenceNumber'] = merchantOrderReferenceNumber
        data['urlRedirect'] = redirect_url
        data['urlNotify'] = notify_url
        if kassenzeichen:
            data['kassenzeichen'] = kassenzeichen


        giropay_transaction.save()

        response = self.call_api(url=self.transaction_start_url, data=data)

        response_hash = response.headers.get('hash')
        response_dict = response.json()
        response_text = response.text

        generated_hash = self._generate_hash_from_text(response_text)

        if response_hash != generated_hash:
            logger.error(_("Response hash {} not compatible with the generated hash {}.").format(response_hash, generated_hash))

        giropay_transaction.reference = response_dict['reference']
        giropay_transaction.latest_response_code = int(response_dict['rc'])
        giropay_transaction.latest_response_msg = response_dict['msg']
        giropay_transaction.save()

        if giropay_transaction.latest_response_code != 0:
            logger.error(_("Transaction Start Response code is {} {}.").format(giropay_transaction.latest_response_code, giropay_transaction.latest_response_msg))
        else:
            giropay_transaction.reference = response_dict['reference']
            giropay_transaction.external_payment_url = response_dict['redirect']
            giropay_transaction.save()
        return giropay_transaction

    def call_api(self, url=None, data=None):
        if not self.auth:
            return False
        if not url.lower().startswith('http'):
            url = '{0}{1}'.format(self.api_url, url)

        generated_hash = self._generate_hash_from_dict(data)
        data.update({'hash': generated_hash})

        try:
            response = requests.post(url, data=data)
        except requests.HTTPError as e:
            logger = logging.getLogger(__name__)
            if hasattr(e, 'errno'):
                logger.error("GiroPay Error {0}({1})".format(e.errno, e.strerror))  # TODO fix error handling according to requests lib
            else:
                logger.error("GiroPay Error({0})".format(e.strerror))  # TODO fix error handling according to requests lib
        else:
            return response
        return False

    def _generate_hash_from_dict(self, data_dict):
        import hashlib
        import hmac
        data_string = "".join([str(val) for val in data_dict.values()])
        data_hash = hmac.new(self.auth['PROJECT_PASSWORD'], "{}".format(data_string).encode("utf-8"), hashlib.md5).hexdigest()
        return data_hash

    def _generate_hash_from_text(self, data_text):
        import hashlib
        import hmac
        data_hash = hmac.new(self.auth['PROJECT_PASSWORD'], data_text.encode("utf-8"), hashlib.md5).hexdigest()
        return data_hash



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-12-20 14:46
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='GiropayTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created_at')),
                ('reference', models.TextField(null=True, verbose_name='reference')),
                ('backend_tx_id', models.TextField(null=True, verbose_name='backend tx id')),
                ('merchant_tx_id', models.CharField(max_length=255, unique=True, verbose_name='merchant tx id')),
                ('merchant_id', models.IntegerField(verbose_name='amount in Cents')),
                ('project_id', models.IntegerField(verbose_name='amount in Cents')),
                ('amount', models.PositiveIntegerField(verbose_name='amount in Cents')),
                ('currency', models.CharField(default='EUR', max_length=3, verbose_name='Currency Code (3 Chars)')),
                ('purpose', models.CharField(max_length=27, verbose_name='purpose')),
                ('iban', models.CharField(blank=True, max_length=34, null=True, verbose_name='iban')),
                ('bic', models.CharField(blank=True, max_length=11, null=True, verbose_name='bic')),
                ('redirect_url', models.TextField(verbose_name='redirect url')),
                ('notify_url', models.TextField(verbose_name='notify url')),
                ('external_payment_url', models.TextField(blank=True, null=True, verbose_name='redirect url to users online banking')),
                ('success_url', models.TextField(verbose_name='success url')),
                ('error_url', models.TextField(verbose_name='error url')),
                ('info_1_label', models.CharField(blank=True, max_length=30, null=True, verbose_name='info 1 label')),
                ('info_1_text', models.CharField(blank=True, max_length=80, null=True, verbose_name='info 1 text')),
                ('info_2_label', models.CharField(blank=True, max_length=30, null=True, verbose_name='info 2 label')),
                ('info_2_text', models.CharField(blank=True, max_length=80, null=True, verbose_name='info 2 text')),
                ('info_3_label', models.CharField(blank=True, max_length=30, null=True, verbose_name='info 3 label')),
                ('info_3_text', models.CharField(blank=True, max_length=80, null=True, verbose_name='info 3 text')),
                ('info_4_label', models.CharField(blank=True, max_length=30, null=True, verbose_name='info 4 label')),
                ('info_4_text', models.CharField(blank=True, max_length=80, null=True, verbose_name='info 4 text')),
                ('info_5_label', models.CharField(blank=True, max_length=30, null=True, verbose_name='info 5 label')),
                ('info_5_text', models.CharField(blank=True, max_length=80, null=True, verbose_name='info 5 text')),
                ('result_payment', models.IntegerField(null=True, verbose_name='return code from giropay transaction')),
                ('result_avs', models.IntegerField(blank=True, null=True, verbose_name='return code of giropay age verification')),
                ('obv_name', models.TextField(blank=True, null=True, verbose_name='obv name')),
                ('latest_response_code', models.IntegerField(null=True, verbose_name='rc field from giropay response')),
                ('latest_response_msg', models.TextField(blank=True, null=True, verbose_name='msg field from giropay response')),
            ],
            options={
                'verbose_name': 'Giropay Transaction',
                'verbose_name_plural': 'Giropay Transaction',
            },
        ),
    ]



from django.utils.translation import gettext as _


RESULT_PAYMENT_STATUS = {
    4000: _("4000 Transaction successful"),
    4001: _("4001 Bank offline"),
    4002: _("4002 Online banking account invalid"),
    4500: _("4500 Unknown payment result"),
    4051: _("4051 Invalid bank account"),
    4101: _("4101 Issuing country invalid or unknown"),
    4102: _("4102 3D-Secure or MasterCard SecureCode authorization failed"),
    4103: _("4103 Validation date of card exceeded"),
    4104: _("4104 Invalid or unknown card type"),
    4105: _("4105 Limited-use card"),
    4106: _("4106 Invalid pseudo-cardnumber"),
    4107: _("4107 Card stolen, suspicious or marked to move in"),
    4108: _("4108 Telephone approval"),
    4151: _("4151 Invalid PayPal token"),
    4152: _("4152 Post-processing necessary at PayPal"),
    4153: _("4153 Change PayPal payment method"),
    4154: _("4154 PayPal-payment is not completed"),
    4501: _("4501 Timeout / no user input"),
    4502: _("4502 User aborted"),
    4503: _("4503 Duplicate transaction"),
    4504: _("4504 Uspicion of manipulation or payment method temporarily blocked"),
    4505: _("4505 Payment method blocked or rejected"),
    4506: _("4506 Invalid Blue Code barcode"),
    4900: _("4900 Transaction rejected"),
}

AGE_VERIFICATION_STATUS = {
    4020: _("4020 Age verification successful"),
    4021: _("4021 Age verification not possible"),
    4022: _("4022 Age verification unsuccessful"),
}

class SHOPPING_CART_TYPE:
    PHYSICAL = 'PHYSICAL'
    DIGITAL = 'DIGITAL'
    MIXED = 'MIXED'
    ANONYMOUS_DONATION = 'ANONYMOUS_DONATION'
    AUTHORITIES_PAYMENT = 'AUTHORITIES_PAYMENT'



from __future__ import unicode_literals

import logging

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _



@python_2_unicode_compatible
class GiropayTransaction(models.Model):
    created_at = models.DateTimeField(_("created_at"), auto_now_add=True)

    reference = models.TextField(_("reference"), null=True)
    backend_tx_id = models.TextField(_("backend tx id"), null=True)
    merchant_tx_id = models.CharField(_("merchant tx id"), max_length=255, unique=True)

    merchant_id = models.IntegerField(_("amount in Cents"))
    project_id = models.IntegerField(_("amount in Cents"))

    amount = models.PositiveIntegerField(_("amount in Cents"))
    currency = models.CharField(_("Currency Code (3 Chars)"), max_length=3, default='EUR')
    purpose = models.CharField(_("purpose"), max_length=27)

    iban = models.CharField(_("iban"), max_length=34, null=True, blank=True)
    bic = models.CharField(_("bic"), max_length=11, null=True, blank=True)

    redirect_url = models.TextField(_("redirect url"))
    notify_url = models.TextField(_("notify url"))

    external_payment_url = models.TextField(_("redirect url to users online banking"), null=True, blank=True)

    success_url = models.TextField(_("success url"))
    error_url = models.TextField(_("error url"))

    info_1_label = models.CharField(_("info 1 label"), max_length=30, null=True, blank=True)
    info_1_text = models.CharField(_("info 1 text"), max_length=80, null=True, blank=True)

    info_2_label = models.CharField(_("info 2 label"), max_length=30, null=True, blank=True)
    info_2_text = models.CharField(_("info 2 text"), max_length=80, null=True, blank=True)

    info_3_label = models.CharField(_("info 3 label"), max_length=30, null=True, blank=True)
    info_3_text = models.CharField(_("info 3 text"), max_length=80, null=True, blank=True)

    info_4_label = models.CharField(_("info 4 label"), max_length=30, null=True, blank=True)
    info_4_text = models.CharField(_("info 4 text"), max_length=80, null=True, blank=True)

    info_5_label = models.CharField(_("info 5 label"), max_length=30, null=True, blank=True)
    info_5_text = models.CharField(_("info 5 text"), max_length=80, null=True, blank=True)

    result_payment = models.IntegerField(_("return code from giropay transaction"), null=True)
    result_avs = models.IntegerField(_("return code of giropay age verification"), null=True, blank=True)  # giropay-ID
    obv_name = models.TextField(_("obv name"), null=True, blank=True)  # giropay-ID

    latest_response_code = models.IntegerField(_("rc field from giropay response"), null=True)  # returned by giropay when calling their API
    latest_response_msg = models.TextField(_("msg field from giropay response"), blank=True, null=True)  # empty when no error

    objects = models.Manager()

    def __str__(self):
        return self.merchant_tx_id

    class Meta:
        verbose_name = _("Giropay Transaction")
        verbose_name_plural = _("Giropay Transaction")



from django.conf.urls import url

from .views import NotifyGiropayView, GiropayReturnView

urlpatterns = [
    url(r'^notify/$', NotifyGiropayView.as_view(), name='notifiy'),
    url(r'^return/$', GiropayReturnView.as_view(), name='return'),
]



import logging
from collections import OrderedDict

from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View, RedirectView

from django_giropay.constants import RESULT_PAYMENT_STATUS
from django_giropay.models import GiropayTransaction

from django_giropay.wrappers import GiropayWrapper
from django_giropay import settings as django_giropay_settings

logger = logging.getLogger(__name__)


def validate_giropay_get_params(giropay_wrapper, get_params):
    desired_variables = ['gcReference', 'gcMerchantTxId', 'gcBackendTxId', 'gcAmount', 'gcCurrency', 'gcResultPayment', 'gcHash']

    # check for expected parameters
    if not all([var in get_params for var in desired_variables]):
        logger.error(
            _("Not all desired variables where part of the GiroPay Notification. Payload: {}").format(str(get_params))
        )
        return False

    # calculate hash and compare
    notification = OrderedDict()
    notification_hash = ""
    for key, value in get_params.items():
        if key == 'gcHash':
            notification_hash = value
        else:
            if key.startswith('gc'):
                notification[key] = value
    generated_hash = giropay_wrapper._generate_hash_from_dict(notification)

    if generated_hash != notification_hash:
        logger.error(
            _("Generated hash does not match received hash. Generated hash: {} \nReceived hash: {}").format(generated_hash, notification_hash)
        )
        return False

    return True


class NotifyGiropayView(View):
    giropay_wrapper = GiropayWrapper()

    def get(self, request, *args, **kwargs):
        get_params = OrderedDict()
        # creating OrderedDict out of query string, because we need it ordered for the hash check
        for query_param in request.META['QUERY_STRING'].split('&'):
            get_params[query_param.split('=')[0]] = "=".join(query_param.split('=')[1:])

        if not validate_giropay_get_params(self.giropay_wrapper, get_params):
            return HttpResponse(status=400)

        try:
            giropay_transaction = GiropayTransaction.objects.get(reference=get_params['gcReference'])
        except GiropayTransaction.DoesNotExist:
            return HttpResponse(status=400)

        giropay_transaction.result_payment = int(get_params['gcResultPayment'])
        giropay_transaction.backend_tx_id = get_params['gcBackendTxId']
        giropay_transaction.save()

        return self.handle_updated_transaction(giropay_transaction=giropay_transaction)

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(NotifyGiropayView, self).dispatch(request, *args, **kwargs)

    def handle_updated_transaction(self, giropay_transaction, expected_statuses=django_giropay_settings.GIROPAY_VALID_TRANSACTION_STATUSES):
        """
            Override to use the giropay_transaction in the way you want.
        """
        if giropay_transaction.result_payment not in expected_statuses:
            logger.error(
                _("GiroPay Result faulty: {}").format(RESULT_PAYMENT_STATUS[giropay_transaction.result_payment] if giropay_transaction.result_payment in RESULT_PAYMENT_STATUS else giropay_transaction.result_payment)
            )
            return HttpResponse(status=400)
        return HttpResponse(status=200)


class GiropayReturnView(RedirectView):
    giropay_wrapper = GiropayWrapper()

    def get_error_url(self):
        return django_giropay_settings.GIROPAY_ERROR_URL

    def get_cancel_url(self, giropay_transaction):
        return giropay_transaction.error_url

    def get_success_url(self, giropay_transaction):
        return giropay_transaction.success_url

    def get_redirect_url(self, *args, **kwargs):
        # creating OrderedDict out of query string, because we need it ordered for the hash check
        get_params = OrderedDict()
        for query_param in self.request.META['QUERY_STRING'].split('&'):
            get_params[query_param.split('=')[0]] = "=".join(query_param.split('=')[1:])

        if not validate_giropay_get_params(self.giropay_wrapper, get_params):
            return self.get_error_url()

        try:
            giropay_transaction = GiropayTransaction.objects.get(reference=get_params['gcReference'])
        except GiropayTransaction.DoesNotExist:
            logger.error(
                _("GiropayTransaction with reference {} does not exist.").format(get_params['gcReference'])
            )
            return self.get_error_url()

        giropay_transaction.result_payment = int(get_params['gcResultPayment'])
        giropay_transaction.backend_tx_id = get_params['gcBackendTxId']
        giropay_transaction.save()

        if giropay_transaction.result_payment not in django_giropay_settings.GIROPAY_VALID_TRANSACTION_STATUSES:
            return self.get_cancel_url(giropay_transaction)
        return self.get_success_url(giropay_transaction)



from django.contrib import admin

from .models import GiropayTransaction


class GiropayTransactionAdmin(admin.ModelAdmin):
    list_display = ('merchant_tx_id', 'reference', 'latest_response_code')
    list_filter = ('latest_response_code',)
    ordering = ('-created_at',)
    fields = ('merchant_tx_id', 'reference', 'latest_response_code')

admin.site.register(GiropayTransaction, GiropayTransactionAdmin)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import re
import os
import sys


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.match("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if 'tests' not in dirnames and not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}

REQUIREMENTS = [
    'Django>=1.11',
    'six>=1.10.0',
]

version = get_version('django_giropay')


if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist upload")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()


setup(
    name='django-giropay',
    author='Particulate Solutions GmbH',
    author_email='tech@particulate.me',
    description=u'Django integration of Giropay.de',
    version=version,
    url='https://github.com/ParticulateSolutions/django-giropay',
    packages=get_packages('django_giropay'),
    package_data=get_package_data('django_giropay'),
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Framework :: Django',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    install_requires=REQUIREMENTS,
    zip_safe=False)



from django.conf import settings

from django_giropay.__init__ import __version__

DJANGO_GIROPAY_VERSION = __version__

GIROPAY_MERCHANT_ID = getattr(settings, 'GIROPAY_MERCHANT_ID', "1234567")
GIROPAY_PROJECT_ID = getattr(settings, 'GIROPAY_PROJECT_ID', "1234")
GIROPAY_PROJECT_PASSWORD = getattr(settings, 'GIROPAY_PROJECT_PASSWORD', b'secure')

GIROPAY_VALID_TRANSACTION_STATUSES = getattr(settings, 'GIROPAY_VALID_TRANSACTION_STATUSES', [4000])

GIROPAY_API_URL = getattr(settings, 'GIROPAY_API_URL', 'https://payment.girosolution.de')

GIROPAY_BANKSTATUS_URL = getattr(settings, 'GIROPAY_BANKSTATUS_URL', '/girocheckout/api/v2/giropay/bankstatus')
GIROPAY_TRANSACTION_START_URL = getattr(settings, 'GIROPAY_TRANSACTION_START_URL', '/girocheckout/api/v2/transaction/start')
GIROPAY_ISSUER_URL = getattr(settings, 'GIROPAY_ISSUER_URL', '/girocheckout/api/v2/giropay/issuer')

# checkout urls
GIROPAY_RETURN_URL = getattr(settings, 'GIROPAY_RETURN_URL', '/giropay/return/')
GIROPAY_SUCCESS_URL = getattr(settings, 'GIROPAY_SUCCESS_URL', '/')
GIROPAY_ERROR_URL = getattr(settings, 'GIROPAY_ERROR_URL', '/')
GIROPAY_CANCELLATION_URL = getattr(settings, 'GIROPAY_CANCELLATION_URL', '/')
GIROPAY_NOTIFICATION_URL = getattr(settings, 'GIROPAY_NOTIFICATION_URL', '/giropay/notify/')

if getattr(settings, 'GIROPAY', False):
    GIROPAY_ROOT_URL = settings.GIROPAY_ROOT_URL



from django_giropay import settings as django_giropay_settings


def build_giropay_full_uri(url):
    if url.startswith('/'):
        url = '{0}{1}'.format(django_giropay_settings.GIROPAY_ROOT_URL, url)
    return url
